"""
This module contains the command line interface for snakeviz.

"""

import argparse
import os
import random
import sys
import threading
import webbrowser
from pstats import Stats
from urllib.parse import quote

from snakeviz import VERSION


# As seen in IPython:
# https://github.com/ipython/ipython/blob/8be7f9abd97eafb493817371d70101d28640919c/IPython/html/notebookapp.py
# See the IPython license at:
# https://github.com/ipython/ipython/blob/master/COPYING.rst.
def random_ports(port, n):
    """Generate a list of n random ports near the given port.
    The first 5 ports will be sequential, and the remaining n-5 will be
    randomly selected in the range [port-2*n, port+2*n].
    """
    for i in range(min(5, n)):
        yield port + i
    for i in range(n-5):
        yield max(1, port + random.randint(-2*n, 2*n))


class SVArgumentParser(argparse.ArgumentParser):
    def error(self, message):
        message = message + '\n\n' + self.format_help()
        args = {'prog': self.prog, 'message': message}
        self.exit(2, '%(prog)s: error: %(message)s' % args)


def build_parser():
    parser = SVArgumentParser(
        description='Start SnakeViz to view a Python profile.')

    parser.add_argument('filename', help='Python profile to view')

    parser.add_argument('-v', '--version', action='version',
                        version=('%(prog)s ' + VERSION))

    parser.add_argument('-H', '--hostname', metavar='ADDR', default='127.0.0.1',
                        help='hostname to bind to (default: %(default)s)')

    parser.add_argument('-p', '--port', type=int, metavar='PORT', default=8080,
                        help='port to bind to; if this port is already in use a '
                             'free port will be selected automatically '
                             '(default: %(default)s)')

    parser.add_argument('-b', '--browser', metavar='BROWSER_PATH',
                        help='name of webbrowser to launch as described in '
                             'the documentation of Python\'s webbrowser module: '
                             'https://docs.python.org/3/library/webbrowser.html')

    parser.add_argument('-s', '--server', action="store_true", default=False,
                        help='start SnakeViz in server-only mode--'
                             'no attempt will be made to open a browser')

    return parser


def main(argv=None):
    parser = build_parser()
    args = parser.parse_args(argv)

    if args.browser and args.server:
        parser.error("options --browser and --server are mutually exclusive")

    filename = os.path.abspath(args.filename)
    if not os.path.exists(filename):
        parser.error('the path %s does not exist' % filename)

    if not os.path.isdir(filename):
        try:
            open(filename)
        except OSError as e:
            parser.error('the file %s could not be opened: %s'
                         % (filename, str(e)))

        try:
            Stats(filename)
        except Exception:
            parser.error(('The file %s is not a valid profile. ' % filename) +
                         'Generate profiles using: \n\n'
                         '\tpython -m cProfile -o my_program.prof my_program.py\n\n'
                         'Note that snakeviz must be run under the same '
                         'version of Python as was used to create the profile.\n')

    filename = quote(filename, safe='')

    hostname = args.hostname
    port = args.port

    if not 0 <= port <= 65535:
        parser.error('invalid port number %d: use a port between 0 and 65535'
                     % port)

    # Go ahead and import the tornado app and start it; we do an inline import
    # here to avoid the extra overhead when just running the cli for --help and
    # the like

    from .main import app
    import tornado.ioloop

    # As seen in IPython:
    # https://github.com/ipython/ipython/blob/8be7f9abd97eafb493817371d70101d28640919c/IPython/html/notebookapp.py
    # See the IPython license at:
    # https://github.com/ipython/ipython/blob/master/COPYING.rst.
    for p in random_ports(port, 10):
        try:
            app.listen(p, address=hostname)
        except OSError:
            print(f'Port {p} in use, trying another.')
        else:
            port = p
            break
    else:
        print('No available port found.')
        return 1

    url = f"http://{hostname}:{port}/snakeviz/{filename}"
    print('snakeviz web server started on %s:%d; enter Ctrl-C to exit' %
           (hostname, port))
    print(url)

    if not args.server:
        try:
            browser = webbrowser.get(args.browser)
        except webbrowser.Error as e:
            parser.error('no web browser found: %s' % e)

        # Launch the browser in a separate thread to avoid blocking the
        # ioloop from starting
        def bt():
            browser.open(url, new=2)
        threading.Thread(target=bt).start()

    try:
        tornado.ioloop.IOLoop.instance().start()
    except KeyboardInterrupt:
        # TODO: Cheap KeyboardInterrupt handler for now; iPython has some nicer
        # stuff for handling SIGINT and SIGTERM that might be worth borrowing
        tornado.ioloop.IOLoop.instance().stop()
        print('\nBye!')

    return 0


if __name__ == '__main__':
    sys.exit(main())



import cProfile
import os
import glob
import shlex
import subprocess as sp
import tempfile
import time
from contextlib import contextmanager
from urllib.parse import quote

import pytest
import requests

from snakeviz import VERSION


@contextmanager
def snakeviz(fname, port=None):
    if port:
        args = f'snakeviz -s --port {port}'
    else:
        args = 'snakeviz -s'

    args += ' ' + fname

    p = sp.Popen(shlex.split(args))
    # give server time to start up
    time.sleep(3)
    yield
    p.terminate()


@pytest.fixture(scope='module')
def prof(request):
    with tempfile.NamedTemporaryFile() as f:
        fname = f.name
    cProfile.runctx('glob.glob("*")', {}, {'glob': glob}, fname)

    def fin():
        if os.path.exists(fname):
            os.remove(fname)
    request.addfinalizer(fin)

    return fname


@pytest.fixture(scope='module', params=[None, 9999])
def port(request):
    return request.param


def snakeviz_url(path, port):
    return 'http://localhost:{}/snakeviz/{}'.format(
        port or 8080,  # default port for snakeviz
        quote(path))


def test_snakeviz_profile(prof, port):
    url = snakeviz_url(prof, port)

    with snakeviz(prof, port=port):
        result = requests.get(url)

    result.raise_for_status()
    assert 'SnakeViz' in result.text


def test_snakeviz_dir(tmpdir, port):
    tmpdir.join('file.txt').write('contents')
    tmpdir.mkdir('subdir')
    url = snakeviz_url(str(tmpdir), port)

    with snakeviz(str(tmpdir), port=port):
        result = requests.get(url)
    result.raise_for_status()

    assert 'file.txt' in result.text
    assert 'subdir/' in result.text


def test_version():
    vcall = sp.Popen(
        ['snakeviz', '--version'], stdout=sp.PIPE, stderr=sp.PIPE)
    out, err = vcall.communicate()
    # in Python <= 3.3 this comes out on stderr, otherwise on stdout
    assert VERSION in out.decode('utf-8') or \
        VERSION in err.decode('utf-8')




import IPython.testing.globalipapp as testing
import IPython.core.error as error

import pytest
import unittest.mock as mock

import snakeviz.ipymagic


@pytest.fixture(scope="module")
def shell():
    result = testing.get_ipython()
    result.run_line_magic("load_ext", "snakeviz")
    return result

def test_snakeviz_magic_browser_default(shell):
    # Since we use the test shell we need to make sure we try to embed the display by
    # faking that we have an interactive shell:
    with mock.patch.object(snakeviz.ipymagic, "_check_ipynb", return_value=True):

        with mock.patch.object(snakeviz.ipymagic, "display") as mock_display:
            shell.run_line_magic("snakeviz", "print()")
            assert mock_display.called
            html = mock_display.call_args[0][0]._repr_html_()
            assert "\" + document.location.hostname + \"" in html


@pytest.mark.parametrize("config_str,test_str",
    (
        ("-h localhost", "localhost"),
        ("-p 8900", "8900"),
         ("--host=localhost", "localhost"),
         ("--port=8900", "8900"),
         ("-h localhost -p 8900", "localhost:8900"),
    )
)
def test_snakeviz_config(config_str, test_str, shell):
    shell.run_line_magic("snakeviz_config", config_str)
    # Since we use the test shell we need to make sure we try to embed the display by
    # faking that we have an interactive shell:
    with mock.patch.object(snakeviz.ipymagic, "_check_ipynb", return_value=True):
        with mock.patch.object(snakeviz.ipymagic, "display") as mock_display:
            shell.run_line_magic("snakeviz", "print()")
            assert mock_display.called
            html = mock_display.call_args[0][0]._repr_html_()
            assert test_str in html


def test_snakeviz_config_unsupported_option(shell):
    with pytest.raises(error.UsageError, match="option -r not recognized"):
        shell.run_line_magic("snakeviz_config", "-r")
